package avalon.categoryClassVioletParser;

import CoreMDELite.Globals;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import PrologDB.*;
import java.io.File;

public class DBWriter {

    public static String DB_FILE_APPENSION = ".pl";
    String outputFile;
    ArrayList<ClassNode> classNodes;
    ArrayList<ClassEdge> classEdges;
    ArrayList<ClassInterface> classInterfaces;
    ClassViolet violetData;
    int numNodes;
    boolean printedVioletInterfaceExtends;
    boolean printedVioletClassImplements;
    int counter = 0;
    LinkedList<Method> allMethods = null;

    public DBWriter(String outputFile, ArrayList<ClassNode> classNodes,
            ArrayList<ClassEdge> classEdges, ArrayList<ClassInterface> classInterfaces,
            ClassViolet violetData, int numNodes) {
        this.outputFile = outputFile;
        this.classNodes = classNodes;
        this.classEdges = classEdges;
        this.classInterfaces = classInterfaces;
        this.violetData = violetData;
        this.numNodes = numNodes;
        this.printedVioletInterfaceExtends = false;
        this.printedVioletClassImplements = false;
        allMethods = new LinkedList<>();
    }

    public void generateVioletDB() {
        try {
            // Step 1: init
            PrintStream out = new PrintStream(outputFile);
            PrintStream eps = new PrintStream(Globals.conformFile);
            File schemaFile = new File(Globals.appPath + "libpl/meta.schema.pl");

            DBSchema dbs = ReadPrologDB.readSchema(schemaFile, eps);
            DB db = new DB(dbs);
            allMethods = new LinkedList<>();

            // Step 2: write class tuples
            boolean foundStart = false;
            TableSchema cs = dbs.findTableSchema("class");
            Table c = db.getTable("class");
            for (ClassNode node : classNodes) {
                Tuple t = makeClassTuple(node, cs);
                if (t != null) {
                    c.addTuple(t);
                }
                foundStart = foundStart || node.name.equals("Start");
            }
            // DON --  the commented out code requires a Start class and getStarted method
            // See below
//            if (!foundStart) {
//                Main.errout("missing Start class declaration");
//            }

            // Step 3: write method tuples
            TableSchema ms = dbs.findTableSchema("method");
            Table m = db.getTable("method");
            writeMethodTuples(m, ms);

            // Step 4: write path tuples
            TableSchema ps = dbs.findTableSchema("path");
            Table p = db.getTable("path");
            writePathTuples(p, ps);

            //  Step 5: make sure that there are no associations
            for (ClassEdge edge : classEdges) {
                Main.errout("one or more associations were defined -- please remove");
                return;
            }
            db.print(out);
            out.close();
        } catch (IOException e) {
            System.err.println(e.getMessage());
            e.printStackTrace();
        }
    }

    private Tuple makeClassTuple(ClassNode node, TableSchema ts) throws IOException {
        String id = node.getId();
        String name = formatAttrsAndMeths(node.getName().replace("«interface»", "").trim());
        HashMap<String, String> attrs = node.getFields();
        HashMap<String, Method> methods = node.getMethods();
        if (methods != null) {
            allMethods.addAll(methods.values());
        }

        boolean hasError = false;
        String ext = null;  // pacify whiney compiler
        String isTemp = null;
        String domain = null;
        if (attrs != null) {
            ext = attrs.get("ext");
            isTemp = attrs.get("isTemp");
            domain = attrs.get("domain");
        } else {
            Main.errout("missing attribute domain, isTemp, ext in class " + name);
            hasError = true;
        }
        if (ext == null) {
            Main.errout("missing attribute ext (maybe others) in class " + name);
            hasError = true;
        }
        if (domain == null) {
            Main.errout("missing attribute domain (maybe others) in class " + name);
            hasError = true;
        }
        if (isTemp == null) {
            Main.errout("missing attribute isTemp (maybe others) in class " + name);
            hasError = true;
        }
        if (ext.equals("pl") && !domain.toLowerCase().equals(name.toLowerCase())) {
            Main.errout("name of prolog database domain must equal to lowercase name of its class: " + name);
            hasError = true;
        }

        if (hasError) {
            return null;
        }
        Tuple t = new Tuple(ts);
        //"\ntable(class,[id,\"name\",\"domain\",\"ext\",temporary,x,y]).\n"
        t.addColumnValue("id", id).addColumnValue("name", name).addColumnValue("domain", domain);
        t.addColumnValue("ext", ext).addColumnValue("temporary", isTemp);
        t.addColumnValue("x", "" + node.getXPos()).addColumnValue("y", "" + node.getYPos());
        return t;
    }

    private void writeMethodTuples(Table tab, TableSchema ts) {
        int counter = 0;
//        
//        out.format("\ntable(method,[id,\"cid\",\"name\",\"resulttype\",\"sig\",\"filename\",ext]).\n");

        boolean getStartedFound = false;
        for (Method meth : allMethods) {
            getStartedFound = getStartedFound || meth.name.equals("getStarted");
            tab.addTuple(meth.makeTuple(ts, counter++));     //meth.printTuple(out, counter++);
        }

        // DON --  the commented out code requires a Start class and getStarted method
        // See above
//        if (!getStartedFound) {
//            Main.errout("method Start.getStarted() not defined");
//        }
    }

    private void writePathTuples(Table tab, TableSchema ts) {
        boolean hasPath = false;
        for (String pathName : Importer.pathInfo.keySet()) {
            hasPath = true;
            String path = Importer.pathInfo.get(pathName);
            // check all paths so that they make sense
            String[] segments = path.replace(" ", "").split("\\.");
            for (String step : segments) {
                if (step.equals("]")) {
                    continue;
                }
                if (step.equals("#")) {
                    continue;
                }
                boolean patherror = true;
                for (Method m : allMethods) {
                    if (m.name.equals(step.replace("[", ""))) {
                        patherror = false;
                        break;
                    }
                }
                if (patherror) {
                    Main.errout("unknown step '" + step + "' in path " + pathName);
                    continue;
                }
            }

            Tuple t = new Tuple(ts);
            t.addColumnValue("name", pathName).addColumnValue("path", path);
            tab.addTuple(t);
        }
        if (!hasPath) {
            Main.errout("no paths specification in Avalon input");
        }
    }

    private static String formatAttrsAndMeths(String input) {
        if (input == null) {
            return "";
        }
        return input;
    }
}
