package avalon.categoryVioletParser;

import CoreMDELite.ConformanceError;
import CoreMDELite.Globals;
import PrologDB.*;
import PrologDB.ReadPrologDB;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Map;
import javax.xml.parsers.ParserConfigurationException;
import org.xml.sax.SAXException;

/**
 *
 * @author dsb
 */
public class Main {

    private static PrintStream out;
    private static PrintStream eps;
    private static boolean hasErrors;
    private static Map<String, String> domainInfo;
    private static Map<String, ArrowData> arrowInfo;
    private static Map<String, String> pathInfo;

    private static class ArrowData {

        String type;
        String exe;

        ArrowData(String type, String exe) {
            this.type = type;
            this.exe = exe;
        }
    }

    /**
     * @param args the command line arguments
     */
    public static void main(String... args) {
        // Step 0: initialize static members
        out = null;
        eps = null;
        hasErrors = false;
        domainInfo = new HashMap<>();
        arrowInfo = new HashMap<>();
        pathInfo = new HashMap<>();

        // Step 1: incorrect arguments prints the marquee.
        if (args.length != 2) {
            System.err.format("Usage: %s <in.state.violet> <out.meta.pl>\n", Main.class.getName());
            System.exit(1);
        }

        // Step 2: create the name of the output prolog file and create its printstream
        String inputFileName = args[0];
        String outputPrologFileName = args[1];
        try {
            out = new PrintStream(new File(outputPrologFileName));
            File epsf = new File(CoreMDELite.Globals.conformFile);
            epsf.delete();
            eps = new PrintStream(epsf);
        } catch (FileNotFoundException e) {
            e.printStackTrace();
            return;
        }
        hasErrors = false;

        // Step 3: read violet's xml file and produce the prolog file
        try {
            File fXmlFile = new File(args[0]);
            DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
            DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
            Document doc = dBuilder.parse(fXmlFile);

            doctype(doc, "com.horstmann.violet.StateDiagramGraph");  // this is a test that confirms we have a violet state diagram

            //optional, but recommended
            //read this - http://stackoverflow.com/questions/13786607/normalization-in-dom-parsing-with-java-how-does-it-work
            doc.getDocumentElement().normalize();
            NodeList voidList = doc.getElementsByTagName("void");

            //ps.format("dbase(metadb,[path,domain,arrow]).\n\n");
            File schemaFile = new File(Globals.appPath + "libpl/meta.schema.pl");

            DBSchema dbs = ReadPrologDB.readSchema(schemaFile, eps);
            DB db = new DB(dbs);

            extractClassTuples(voidList, dbs, db);
            extractMethodTuples(voidList, dbs, db);
            extractPath(voidList, dbs, db);

            db.print(out);
            out.close();

        } catch (ParserConfigurationException | SAXException | IOException e) {
            eps.print(e.getMessage());
            e.printStackTrace();
            hasErrors = true;
        }
        eps.close();
        out.close();
        if (hasErrors) {
            throw new ConformanceError("Conformance errors in parsing " + inputFileName + ". see " + Globals.conformFile);
        }
    }

    public static void extractMethodTuples(NodeList voidList, DBSchema dbs, DB db) {
        int counter = 0;
        TableSchema ts = dbs.findTableSchema("method");
        Table tab = db.getTable("method");
        Table ctab = db.getTable("class");

        // Step 1: let's plow through all of the <void> nodes to find <void method="connect">
        //         instances
        for (int i = 0; i < voidList.getLength(); i++) {
            Node voidNode = voidList.item(i);
            Element evoidNode = (Element) voidNode;
            String cls = evoidNode.getAttribute("method");
            if (!cls.equals("connect")) {
                continue;
            }

            // Step 1.1: now lets get the starting and ending domain names.  these are found in 
            //           <object idref=""> nodes -- there are only 3 of them, and they are the last 2.
            //           start and end are the domain identifiers
            NodeList objects = evoidNode.getElementsByTagName("object");
            if (objects.getLength() != 3) {
                eps.format("Internal Error: can't determine domain names\n");
                hasErrors = true;
                continue;
            }

            String start = ((Element) objects.item(1)).getAttribute("idref").toLowerCase();
            String end = ((Element) objects.item(2)).getAttribute("idref").toLowerCase();

            // Step 1.2: we found a <void method="connect"> node.  Now, we need to find the
            //           user given name of that node.  we need to find the <string> children
            //           there should only be one and extract the name in variable arrowName
            //           note: arrowName can be compound a1,a2,a3 -- we will unravel this later
            NodeList strings = evoidNode.getElementsByTagName("string");
            if (strings.getLength() != 1) {
                eps.format("arrow from id %s to id %s should have a name -- see .meta.pl output\n", start, end);
                hasErrors = true;
                continue;
            }
            Element estrings0 = (Element) strings.item(0);
            String arrowName = estrings0.getFirstChild().getNodeValue();
            arrowName = arrowName.replaceAll("\\r|\\n", "");

            // Step 1.3: output an arrow tuple -- unravel arrowName if it is compound
            String[] pieces = arrowName.split(";");
            for (String arrowdef : pieces) {

                String[] parts = arrowdef.split("=");
                if (parts.length != 2) {
                    eps.format("arrow (%s) has incorrect name=impl format\n", arrowdef);
                    hasErrors = true;
                    continue;
                }

                // part[0] has form methodname [(arglist)] [arraybracket]
                String[] mparts = parts[0].split("\\(");
                String methodName, sig;
                String hasBrackets = "";
                switch (mparts.length) {
                    case 1:
                        methodName = mparts[0].trim();
                        sig = "";
                        break;
                    case 2:
                        methodName = mparts[0].trim();
                        sig = mparts[1].replace(" ", "").replace(")", "").trim();
                        break;
                    case 3:
                        methodName = mparts[0].trim() + "[]";
                        sig = mparts[1].replace(" ", "");
                        break;
                    default:
                        eps.format("method signature (%s) has incorrect name(arglist)arraybracket format\n", parts[0]);
                        hasErrors = true;
                        continue;
                }
                if (methodName.endsWith("[]")) {
                    hasBrackets = "[]";
                    methodName = methodName.replace("[]", "");
                }

                // part[1] has form <blah>.ext
                int indx = parts[1].lastIndexOf('.');
                String ext = parts[1].substring(indx + 1).trim();
                String filename = parts[1].substring(0, indx).trim();
                String resulttype = ctab.findFirst("id", end).get("name");

                if (!(ext.equals("pl") || ext.equals("java") || ext.equals("vm"))) {
                    eps.println("extension `" + ext + "' must be pl or java or vm or PL");
                    hasErrors = true;
                }

                Tuple t = new Tuple(ts);
                t.addColumnValue("id", "id" + counter++).addColumnValue("cid", start).addColumnValue("name", methodName);
                t.addColumnValue("resulttype", resulttype + hasBrackets).addColumnValue("sig", sig).addColumnValue("filename", filename);
                t.addColumnValue("ext", ext);
                tab.addTuple(t);
            }

        }

        // Now check to see if all paths reference legit arrows -- and simply report errors
        for (String k : pathInfo.keySet()) {
            String path = pathInfo.get(k);
            String[] pathList = path.split(".");
            for (String pl : pathList) {
                if (!pl.equals("[") && arrowInfo.get(pl) == null && pathInfo.get(pl) == null) {
                    eps.format("unrecognizable step %s in path %s\n", pl, k);
                    hasErrors = true;
                }
            }
        }

    }

    public static void extractClassTuples(NodeList voidList, DBSchema dbs, DB db) {

        TableSchema cs = dbs.findTableSchema("class");
        Table c = db.getTable("class");

        for (int i = 0; i < voidList.getLength(); i++) {
            Node voidNode = voidList.item(i);
            Element evoidNode = (Element) voidNode;
            String cls = evoidNode.getAttribute("method");
            if (!cls.equals("addNode")) {
                continue;
            }

            // Step 2.1: further qualification.  inside the <void method="addNode">
            //           we have to figure out what node is being added.  So we have
            //           to find all <object/> nodes -- really just the first --
            //           to determine if the object to be added is a StateNode or not.
            //           we are only interested in StateNodes. The result of this
            //           code is obtaining the identifier of the domain (in variable id).
            NodeList objects = evoidNode.getElementsByTagName("object");
            Element eobject0 = (Element) objects.item(0);
            String className = eobject0.getAttribute("class");
            if (!className.endsWith(".StateNode")) // could be another kind of node, as in NoteNode
            {
                continue;
            }
            String id = eobject0.getAttribute("id").toLowerCase();

            // Step 2.1: Now that we know we are adding a domain, we have to get its listed name
            //           we find the first <string> node and extract its value.  That's the user-
            //           given name for the domain (in string arrowName).
            NodeList strings = evoidNode.getElementsByTagName("string");
            if (strings.getLength() != 3) {
                eps.format("Internal Error: incorrect number of <string> elements %s\n", strings.getLength()); // if this fails, the xml format has changed
                hasErrors = true;
                continue;
            }

            String[] parts;
            Element estrings0 = (Element) strings.item(0);
            String nodeName = estrings0.getFirstChild().getNodeValue();
            nodeName = nodeName.replaceAll("\\r|\\n", "");
            parts = nodeName.split(";");

            // Check -- should have 4 parts
            if (parts.length != 4) {
                eps.format("node " + nodeName + " is not in the correct format name;domain=d;ext=e;isTemp=f;\n");
                hasErrors = true;
                continue;
            }
            String name = parts[0];
            String domaindecl = parts[1];
            String extdecl = parts[2];
            String isTempdecl = parts[3];

            // now extract the parts that are important
            String dom = extractFieldValue("domain", domaindecl);
            String ext = extractFieldValue("ext", extdecl);
            String tmp = extractFieldValue("isTemp", isTempdecl);

            // Step 2.3: last, we need to extract the violet (x,y) position of the domain.
            //           we find all <void method="setLocation">  nodes -- this should be
            //           the last node #7
            //           
            NodeList voids = evoidNode.getElementsByTagName("void");
            if (voids.getLength() != 7) {
                eps.format("Internal Error: incorrect number of <void> elements %d\n", voids.getLength()); // if this fails, the xml format has changed
                hasErrors = true;
                continue;
            }
            Element eee = (Element) voids.item(6);
            cls = eee.getAttribute("method");
            if (!cls.equals("setLocation")) {
                eps.format("Internal Error: expected setLocation -- but found: %s\n", cls);
                hasErrors = true;
                continue;
            }

            // Step 2.4: now get the nested <double> nodes -- there are 2.  The first
            //           exposes the x value, the second exposes the y value
            NodeList doubles = eee.getElementsByTagName("double");
            if (doubles.getLength() != 2) {
                eps.format("Internal Error: incorrect number of <double> elements %d\n", doubles.getLength());
                hasErrors = true;
                continue;
            }
            String x = ((Element) doubles.item(0)).getFirstChild().getNodeValue();
            String y = ((Element) doubles.item(1)).getFirstChild().getNodeValue();
            int ix = (Double.valueOf(x)).intValue();
            int iy = (Double.valueOf(y)).intValue();

            if (ext == null) {
                eps.println("missing attribute ext (maybe others) in class " + name);
                hasErrors = true;
            }
            if (dom == null) {
                eps.println("missing attribute domain (maybe others) in class " + name);
                hasErrors = true;
            }
            if (tmp == null) {
                eps.println("missing attribute isTemp (maybe others) in class " + name);
                hasErrors = true;
            }
            if (ext.equals("pl") && !dom.toLowerCase().equals(name.toLowerCase())) {
                eps.println("name of prolog database domain must equal to lowercase name of its class: " + name);
                hasErrors = true;
            }

            Tuple t = new Tuple(cs);
            t.addColumnValue("id", id).addColumnValue("name", name).addColumnValue("domain", dom);
            t.addColumnValue("ext", ext).addColumnValue("temporary", tmp);
            t.addColumnValue("x", "" + ix).addColumnValue("y", "" + iy);

            c.addTuple(t);

        }
    }

    private static String extractFieldValue(String field, String decl) {
        String[] parts = decl.trim().split("=");
        if (parts.length != 2 || !field.equals(parts[0].trim())) {
            eps.format("declaration ( %s ) does not have the correct %s=value format\n", decl, field);
            hasErrors = true;
            return "<error>";
        }
        return parts[1].trim();
    }

    public static void extractPath(NodeList voidList, DBSchema dbs, DB db) {

        boolean hasPaths = false;
        TableSchema ts = dbs.findTableSchema("path");
        Table tab = db.getTable("path");

        // Step 2: let's plow through all of the <void/> nodes with attribute method="addNode"
        for (int i = 0; i < voidList.getLength(); i++) {
            Node voidNode = voidList.item(i);
            Element evoidNode = (Element) voidNode;
            String cls = evoidNode.getAttribute("method");
            if (!cls.equals("addNode")) {
                continue;
            }

            // Step 2.1: further qualification.  inside the <void method="addNode">
            //           we have to figure out what node is being added.  So we have
            //           to find all <object/> nodes -- really just the first --
            //           to determine if the object to be added is a NoteNode or not.
            NodeList objects = evoidNode.getElementsByTagName("object");
            Element eobject0 = (Element) objects.item(0);
            String className = eobject0.getAttribute("class");
            if (!className.endsWith(".NoteNode")) // could be another kind of node, as in NoteNode
            {
                continue;
            }

            // Step 2.2: Now that we know we are adding a note, we have to get its string
            //           we do this by finding the first <string> node and extracting its
            //           value, and then into parts[]
            NodeList strings = evoidNode.getElementsByTagName("string");

            Element estrings0 = (Element) strings.item(0);
            String nodestring = estrings0.getFirstChild().getNodeValue();
            String[] parts = nodestring.split("\\r|\\n");

            // Step 2.3: now decide what kind of note this is: either it is Arrows or Paths note
            if (parts[0].startsWith("Paths")) {

                hasPaths = true;
                // Step 2.3.2: for each line print it out as a tuple
                for (int j = 1; j < parts.length; j++) {
                    String p = parts[j].trim().replace(" ", "");
                    if (p.equals("")) {
                        continue;
                    }
                    if (!p.endsWith(";")) {
                        eps.format("path %s does not end in ';'\n", parts[j]);
                        hasErrors = true;
                    } else {
                        p = p.replace(";", "");
                    }
                    String s[] = p.split("=");
                    if (s.length != 2) {
                        eps.format("path %s does not have the right format", parts[j]);
                        hasErrors = true;
                        continue;
                    }

                    Tuple t = new Tuple(ts);
                    t.addColumnValue("name", s[0]).addColumnValue("path", s[1]);
                    tab.addTuple(t);

                    // used later for path checking
                    pathInfo.put(s[0].trim(), s[1].trim());
                }

            }
            // ignore otherwise
        }
        if (!hasPaths) {
            eps.format("no paths specification in Avalon input");
            hasErrors = true;
        }

    }

    static void doctype(Document doc, String name) {
        NodeList objectList = doc.getElementsByTagName("object");
        String cls = "&&#&&&";
        if (objectList.getLength() != 0) {
            Element node = (Element) objectList.item(0);
            cls = node.getAttribute("class");
        }
        if (!cls.equals(name)) { // means test has failed
            System.err.format("violet document type is not %s\n",name);
            System.exit(1);
        }
    }
}
