package avalon.class2schema;

import CoreMDELite.ConformanceError;
import java.io.PrintStream;
import java.util.ArrayList;
import CoreMDELite.Error;
import CoreMDELite.Globals;
import PrologDB.*;

public class DBWriter {

    DBSchema dbs = null;
    public static String DB_FILE_APPENSION = ".pl";
    String outputFile;
    ArrayList<ClassNode> classNodes;
    ArrayList<ClassEdge> classEdges;
    ArrayList<ClassInterface> classInterfaces;
    ClassViolet violetData;
    int numNodes;
    boolean printedVioletInterfaceExtends;
    boolean printedVioletClassImplements;
    int counter = 0;

    PrintStream out, eps;
    boolean hasErrors;
    String inputFile;

    public DBWriter(String inputFile, String outputFile, ArrayList<ClassNode> classNodes,
            ArrayList<ClassEdge> classEdges, ArrayList<ClassInterface> classInterfaces,
            ClassViolet violetData, int numNodes) {
        this.inputFile = inputFile;
        this.outputFile = makeFirstLowerCase(outputFile);
        this.classNodes = classNodes;
        this.classEdges = classEdges;
        this.classInterfaces = classInterfaces;
        this.violetData = violetData;
        this.numNodes = numNodes;
        this.printedVioletInterfaceExtends = false;
        this.printedVioletClassImplements = false;
    }

    private DBSchema createDBase(String outputFile) {
        // Step 1: extract the tableName of the database.  node.tableName = "something/blah/NAME.schema.pl"\
        String dbaseName = outputFile.replaceAll(".schema.pl", "");
        int i = dbaseName.lastIndexOf("/");
        dbaseName = dbaseName.substring(i + 1);
        return new DBSchema(dbaseName);

    }

    private void createTableDefs() {
        for (ClassNode node : classNodes) {
            TableSchema ts = new TableSchema(node.name);
            for (String a : node.getAttributes()) {
                boolean isQuoted = a.contains("\"");
                if (isQuoted) {
                    a = a.replace("\"", "");
                }
                Column c = new Column(a, isQuoted);
                ts.addColumn(c);
            }
            dbs.addTableSchema(ts);
        }
    }

    public void generateSchema() {
        try {
            // Step 1: open schema file
            out = new PrintStream(outputFile);
            eps = new PrintStream(Globals.errorFile);
            hasErrors = false;

            // Step 2: convert associations into attributes
            addAssociationAttributes();

            // Step 3: add inheritance links
            addInheritanceLinks();

//            // Step 4: propagate inherited attributes
//            propagateAttributes();
            // Step 6: generate the schema
            dbs = createDBase(outputFile);
            createTableDefs();
            createSubTableDefs();
            dbs.flatten();

            dbs.printFullSchema(out);
//            createTableDefs();
//            writeTupleDefs();
//            writeInheritanceRules();

            out.close();
            eps.close();
        } catch (Exception e) {
            throw new Error(e.getMessage());
        }
        if (hasErrors) 
            throw new ConformanceError("errors detected in parsing file "+inputFile);
    }

    private void createSubTableDefs() {
        for (ClassNode n : classNodes) {
            TableSchema nts = dbs.findTableSchema(makeFirstLowerCase(n.name));
            if (!n.parentId.equals("")) {
                ClassNode p = findNode(n.parentId);
                TableSchema pts = dbs.findTableSchema(makeFirstLowerCase(p.name));
                SubTableSchema sts = dbs.findSubTableSchema(pts.getName());
                if (sts == null) {
                    sts = new SubTableSchema(pts);
                    dbs.addSubTableSchema(sts);
                }
                sts.addSubTableSchema(nts);
            }
        }
    }

    private void addAssociationAttributes() {
        ClassNode n;
        String label;

        for (ClassEdge e : classEdges) {
            if (e.startArrowHead.equals("V")) {
                n = findNode(e.endsAtClass);
                label = e.startLabel;
                if (label == null || label.equals("")) {
                    eps.format("no attribute label for arrow-end of association\n");
                    hasErrors = true;
                } else {
                    n.setAttribute(label);
                }
            }
            if (e.endArrowHead.equals("V")) {
                n = findNode(e.startsAtClass);
                label = e.endLabel;
                if (label == null || label.equals("")) {
                    eps.format("no attribute label for arrow-end of association\n");
                    hasErrors = true;
                } else {
                    n.setAttribute(label);
                }
            }
        }
    }

    private void addInheritanceLinks() {
        ClassNode n1, n2;

        for (ClassEdge e : classEdges) {
            if (e.startArrowHead.equals("TRIANGLE")) {
                n1 = findNode(e.startsAtClass);
                n2 = findNode(e.endsAtClass);
                n1.setParentId(n2.id);
            } else if (e.endArrowHead.equals("TRIANGLE")) {
                n1 = findNode(e.startsAtClass);
                n2 = findNode(e.endsAtClass);
                n1.setParentId(n2.id);
            }
        }
    }

    private ClassNode findNode(String name) {
        for (ClassNode node : classNodes) {
            if (node.id.equals(name)) {
                return node;
            }
        }
        throw new Error("cannot find class with name " + name);
    }

    private static String makeFirstLowerCase(String s) {
        boolean quotes = false;
        if (s.startsWith("\"")) {
            s = s.replace("\"", "");
            quotes = true;
        }
        String tail = s.substring(1);
        String head = s.substring(0, 1).toLowerCase();
        String result = head + tail;
        if (quotes) {
            result = "\"" + result + "\"";
        }
        return result;
    }
}
