package avalon.classVioletParser;

import CoreMDELite.Globals;
import PrologDB.*;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;

public class DBWriter {

    public static String DB_FILE_APPENSION = ".pl";
    String outputFile;
    ArrayList<ClassNode> classNodes;
    ArrayList<ClassEdge> classEdges;
    ArrayList<ClassInterface> classInterfaces;
    ClassViolet violetData;
    int numNodes;
    boolean printedVioletInterfaceExtends;
    boolean printedVioletClassImplements;
    int counter = 0;

    public DBWriter(String outputFile, ArrayList<ClassNode> classNodes,
            ArrayList<ClassEdge> classEdges, ArrayList<ClassInterface> classInterfaces,
            ClassViolet violetData, int numNodes) {
        this.outputFile = outputFile.toLowerCase();
        this.classNodes = classNodes;
        this.classEdges = classEdges;
        this.classInterfaces = classInterfaces;
        this.violetData = violetData;
        this.numNodes = numNodes;
        this.printedVioletInterfaceExtends = false;
        this.printedVioletClassImplements = false;
    }

    DB db;
    
    public void generateVioletDB() {
        try {
            PrintStream eps = new PrintStream(new FileOutputStream(Globals.conformFile,true));
            File schemaFile = new File(Globals.appPath + "libpl/violetpl.schema.pl");
            
            DBSchema dbs = ReadPrologDB.readSchema(schemaFile,eps);
            db = new DB(dbs);

            TableSchema cs = dbs.findTableSchema("violetClass");
            Table c = db.getTable("violetClass");
            for (ClassNode node : classNodes) {
                makeClassTuple(node, c, cs);
            }

            TableSchema is = dbs.findTableSchema("violetInterface");
            Table i = db.getTable("violetInterface");
            for (ClassInterface node : classInterfaces) {
                makeInterfaceTuple(node, i, is);
            }

            TableSchema as = dbs.findTableSchema("violetAssociation");
            Table a = db.getTable("violetAssociation");
            for (ClassEdge edge : classEdges) {
                makeAssociationTuple(edge, a, as);
            }
            
            PrintStream out = new PrintStream(outputFile);
            db.print(out);
            out.close();
        } catch (IOException e) {
            System.err.println(e.getMessage());
            e.printStackTrace();
        }
    }

    private static void makeInterfaceTuple(ClassInterface node, Table tab, TableSchema ts) {
        String id = node.getId();
        String name = node.getName();
        // Violet automatically places <<inteface>> as part of the name;
        // this and any embedded blanks should be stripped.

        name = name.replace("«interface» ", "").replace("«interface»", "").replace("interface ","").replace(" ",";").trim();  // Added by *DSB*
        String methods = formatAttrsAndMeths(node.getMethods().toString());
        
        //table(violetInterface,[id,\"name\",\"methods\",x,y]).\n");
        Tuple t = new Tuple(ts);
        t.addColumnValue("id",id).addColumnValue("name", name);
        t.addColumnValue("methods",methods).addColumnValue("x", ""+node.getXPos()).addColumnValue("y", ""+node.getYPos());
        tab.addTuple(t);
    }

    private static void makeClassTuple(ClassNode node, Table tab, TableSchema ts)  {
        String id = node.getId();
        String name = formatAttrsAndMeths(node.getName().replace("«interface»", "").trim());
        String attrs = formatAttrsAndMeths(node.getAttributes().toString());
        String methods = formatAttrsAndMeths(node.getMethods().toString().replace("'",""));

        attrs = attrs.replace("'", "");
        attrs = Importer.allFirstCharsAreLowerCase(attrs, ";");
        
        //table(violetClass,[id,\"name\",\"fields\",\"methods\",x,y]).\n");
        Tuple t = new Tuple(ts);
        t.addColumnValue("id",id).addColumnValue("name", name).addColumnValue("fields", attrs);
        t.addColumnValue("methods",methods).addColumnValue("x", ""+node.getXPos()).addColumnValue("y", ""+node.getYPos());
        tab.addTuple(t);
    }

    private void makeAssociationTuple(ClassEdge edge, Table tab, TableSchema ts)  {
        String startId = edge.getStartsAtClass();
        String endId = edge.getEndsAtClass();
        String startArrow = formatTextVariable(edge.getStartArrowHead());
        String endArrow = formatTextVariable(edge.getEndArrowHead());
        String startLabel = edge.getStartLabel();
        String endLabel = edge.getEndLabel();
        String lineStyle = formatTextVariable(edge.getLineStyle());
        String startsType = edge.getStartsType();
        String endsType = edge.getEndsType();

        startLabel = startLabel.replace(",", ";");
        startLabel = formatTextVariable(Importer.allFirstCharsAreLowerCase(startLabel, ";"));

        //startLabel = formatTextVariable(Importer.allFirstCharsAreLowerCase(startLabel.replace(",", ";"), ";"));
        endLabel = formatTextVariable(Importer.allFirstCharsAreLowerCase(endLabel.replace(",", ";"), ";"));

        //table(violetAssociation,[id,\"role1\",\"arrow1\",type1,\"role2\",\"arrow2\",type2,\"lineStyle\",cid1,cid2]).\n")
        Tuple t = new Tuple(ts);
        t.addColumnValue("id","id" + counter++).addColumnValue("role1", startLabel).addColumnValue("arrow1", startArrow);
        t.addColumnValue("type1",startsType).addColumnValue("role2",endLabel).addColumnValue("arrow2", endArrow);
        t.addColumnValue("type2",endsType).addColumnValue("lineStyle", lineStyle);
        t.addColumnValue("cid1", startId).addColumnValue("cid2", endId);
        tab.addTuple(t);
        
        // this was added by DB: here's the problem.  we don't check to
        // see if associations -- which will eventually add attributes to tables
        // will cause errors in schemas later on.  Like duplicate names.  
        // this is a fundamental problem in Avalon's design.  What I should
        // have done is to convert a violet drawing directly into a schema,
        // and where I could check for errors; the core checks of unique column names
        // and non-cyclic inheritance graphs would be done right then and there.
    }
    
    private static String formatAttrsAndMeths(String input) {
        if (input == null || input.equals("null") || input.equals("[]")) {
            return ""; //''";
        }
        String output = input.replace("[", ""); //'");
        output = output.replace("]", ""); //"'");
        output = output.replace(", ", ";");

        return output;
    }

    private static String formatTextVariable(String input) {
        if (input == null || input.equals("null")) {
            return ""; // ''";
        }
        return input;
    }
}
