package cdexchange.classVioletParser;

import PrologDB.Table;
import PrologDB.ReadPrologDB;
import PrologDB.TableSchema;
import PrologDB.DBSchema;
import PrologDB.Tuple;
import PrologDB.DB;
import CoreMDELite.Globals;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;

public class DBWriter {

    public static String DB_FILE_APPENSION = ".pl";
    String outputFile;
    ArrayList<ClassNode> classNodes;
    ArrayList<ClassEdge> classEdges;
    ArrayList<ClassInterface> classInterfaces;
    ClassViolet violetData;
    int numNodes;
    boolean printedVioletInterfaceExtends;
    boolean printedVioletClassImplements;
    int counter = 0;

    public DBWriter(String outputFile, ArrayList<ClassNode> classNodes,
            ArrayList<ClassEdge> classEdges, ArrayList<ClassInterface> classInterfaces,
            ClassViolet violetData, int numNodes) {
        this.outputFile = outputFile;
        this.classNodes = classNodes;
        this.classEdges = classEdges;
        this.classInterfaces = classInterfaces;
        this.violetData = violetData;
        this.numNodes = numNodes;
        this.printedVioletInterfaceExtends = false;
        this.printedVioletClassImplements = false;
    }

    public void generateVioletDB() {
        try {
            PrintStream eps = new PrintStream(Globals.conformFile);

            File violetSchemaPl = new File(Globals.appPath + "libpl/violetdb.schema.pl");
            DBSchema dbs = ReadPrologDB.readSchema(violetSchemaPl, eps);
            DB db = new DB(dbs);
            
            TableSchema vcs = dbs.findTableSchema("violetClass");
            Table vc = db.getTable("violetClass");
            for( ClassNode node : classNodes) {
                vc.addTuple(makeClassTuple(node,vcs));
            }

            TableSchema vis = dbs.findTableSchema("violetInterface");
            Table vi = db.getTable("violetInterface");
            for (ClassInterface node : classInterfaces) {
                vi.addTuple(makeInterfaceTuple(node,vis));
            }

            TableSchema vas = dbs.findTableSchema("violetAssociation");
            Table va = db.getTable("violetAssociation");
            for (ClassEdge edge : classEdges) {
                va.addTuple(makeAssociationTuple(edge,vas));
            }
            
            TableSchema vmls = dbs.findTableSchema("violetMiddleLabels");
            Table vml = db.getTable("violetMiddleLabels");
            for (ClassEdge edge : classEdges) {
                Tuple t = makeMiddleLabelTuple(edge,vmls);
                if (t!=null)
                    vml.addTuple(t);
            }
            
            PrintStream out = new PrintStream(outputFile);
            db.print(out);
            out.close();
        } catch (IOException e) {
            System.err.println(e.getMessage());
            e.printStackTrace();
        }
    }

    private static Tuple makeInterfaceTuple(ClassInterface node, TableSchema ts) throws IOException {
        Tuple t = new Tuple(ts);
        String id = node.getId();
        String name = node.getName();
        // Violet automatically places <<inteface>> as part of the name;
        // this and any embedded blanks should be stripped.

        name = name.replace("«interface» ", "").replace("«interface»", "").replace("interface ","").replace(" ",";").trim();  // Added by *DSB*
        String methods = formatAttrsAndMeths(node.getMethods().toString());

        t.addColumnValue("id", id).addColumnValue("name", name).addColumnValue("methods", methods);
        t.addColumnValue("x",""+node.getXPos()).addColumnValue("y",""+node.getYPos());
        return t;
    }

    private static Tuple makeClassTuple(ClassNode node, TableSchema ts) throws IOException {
        String id = node.getId();
        String name = formatAttrsAndMeths(node.getName().replace("«interface»", "").trim());
        String attrs = formatAttrsAndMeths(node.getAttributes().toString());
        String methods = formatAttrsAndMeths(node.getMethods().toString());

        attrs = attrs.replace("'", "");
        attrs = Importer.allFirstCharsAreLowerCase(attrs, ";");
        
        Tuple t = new Tuple(ts);
        t.addColumnValue("id", id).addColumnValue("name", name).addColumnValue("fields", attrs);
        t.addColumnValue("methods",methods).addColumnValue("x", ""+node.getXPos()).addColumnValue("y",""+node.getYPos());
        return t;
    }

    private Tuple makeAssociationTuple(ClassEdge edge, TableSchema ts) throws IOException {
        String startId = edge.getStartsAtClass();
        String endId = edge.getEndsAtClass();
        String startArrow = formatTextVariable(edge.getStartArrowHead());
        String endArrow = formatTextVariable(edge.getEndArrowHead());
        String startLabel = edge.getStartLabel();
        String endLabel = edge.getEndLabel();
        String lineStyle = formatTextVariable(edge.getLineStyle());
        String startsType = edge.getStartsType();
        String endsType = edge.getEndsType();

        startLabel = startLabel.replace(",", ";");
        startLabel = formatTextVariable(Importer.allFirstCharsAreLowerCase(startLabel, ";"));
        endLabel = formatTextVariable(Importer.allFirstCharsAreLowerCase(endLabel.replace(",", ";"), ";"));

        //Always need to write an association, regardless of edge type, for Violet purposes
        //violetAssociation,[id,"role1","arrow1",type1,"role2","arrow2",type2,"lineStyle",cid1,cid2]
        Tuple t = new Tuple(ts);
        t.addColumnValue("id","id"+ counter++).addColumnValue("role1", startLabel);
        t.addColumnValue("arrow1",startArrow).addColumnValue("type1",startsType);
        t.addColumnValue("role2",endLabel).addColumnValue("arrow2", endArrow);
        t.addColumnValue("type2", endsType).addColumnValue("lineStyle",lineStyle);
        t.addColumnValue("cid1",startId).addColumnValue("cid2",endId);
        return t;
    }


private Tuple makeMiddleLabelTuple(ClassEdge edge, TableSchema ts) throws IOException {
        String fromId = edge.getStartsAtClass();
        String toId = edge.getEndsAtClass();
        String middleLabel = formatTextVariable(edge.getMiddleLabel());

        if (middleLabel.equals("")) {
            return null;
        }
        
        //table(violetMiddleLabels,[id,cid1,cid2,"label"]).
        Tuple t = new Tuple(ts);
        t.addColumnValue("id", "id"+ counter++).addColumnValue("cid1", fromId);
        t.addColumnValue("cid2",toId).addColumnValue("label", middleLabel);
        return t;
    }

    private static String formatAttrsAndMeths(String input) {
        if (input == null || input.equals("null") || input.equals("[]")) {
            return ""; //''";
        }
        String output = input.replace("[", ""); //'");
        output = output.replace("]", ""); //"'");
        output = output.replace(", ", ";");

        return output;
    }

    private static String formatTextVariable(String input) {
        if (input == null || input.equals("null")) {
            return ""; // ''";
        }
        return input;
    }
}
