/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package PrologDB;

import CoreMDELite.Error;
import CoreMDELite.Globals;
import java.io.File;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.LinkedList;

/**
 *
 * @author dsb
 */
public class DBSchema {

    private String name;
    private LinkedList<TableSchema> schemas;
    private LinkedList<SubTableSchema> subtables;

    public DBSchema(String name, TableSchema... schemas) {
        this.name = name;
        this.schemas = new LinkedList<>();
        this.schemas.addAll(Arrays.asList(schemas));
        subtables = new LinkedList<>();
    }

    public DBSchema(String name) {
        this.name = name;
        this.schemas = new LinkedList<>();
        subtables = new LinkedList<>();
    }

    public void addTableSchema(TableSchema ts) {
        schemas.add(ts);
    }

    public void addSubTableSchema(SubTableSchema s) {
        subtables.add(s);
    }

    public String getName() {
        return name;
    }

    public LinkedList<TableSchema> getTableSchemas() {
        return schemas;
    }

    public int size() {
        return schemas.size();
    }

    public TableSchema findTableSchema(String tableName) {
        for (TableSchema ts : schemas) {
            if (ts.getName().equals(tableName)) {
                return ts;
            }
        }
        throw new Error("cannot find table " + tableName + " in database " + name);
    }

    public SubTableSchema findSubTableSchema(String tableName) {
        for (SubTableSchema s : subtables) {
            if (s.getName().equals(tableName)) {
                return s;
            }
        }
        return null;
    }

    // propagate attributes of supertables to subtables
    public void flatten() {
        // Step 1: must order subtable list by dominance (parent dominates a child)
        LinkedList<SubTableSchema> list = subtables;
        subtables = new LinkedList<>();
        while (!list.isEmpty()) {
            SubTableSchema next = dominant(list);
            list.remove(next);
            subtables.add(next);
        }

        // Step 2: propagate parent attributes to subattributes
        for (SubTableSchema st : subtables) {
            st.flatten();
        }
    }

    private SubTableSchema dominant(LinkedList<SubTableSchema> list) {
        boolean foundit = true;
        SubTableSchema last = null;

        for (SubTableSchema outer : list) {
            last = outer;
            for (SubTableSchema inner : list) {
                if (inner.contains(outer.getSuper())) {
                    foundit = false;
                    break;
                }
            }
            if (foundit) {
                return outer;
            }
            foundit = true;
        }
        throw new Error("table " + last.getName() + " participates in inheritance cycle");
    }

    public String printHeader() {
        String db = "dbase(" + name + ",[";
        String comma = "";
        for (TableSchema ts : schemas) {
            db = db + comma + ts.getName();
            comma = ",";
        }
        return db + "]).\n";
    }

    public void print(PrintStream out) {
        out.format("%s\n", printHeader());

        for (TableSchema ts : schemas) {
            ts.print(out);
        }

        if (!subtables.isEmpty()) {
            out.format("\n");
            for (SubTableSchema st : subtables) {
                st.print(out);
            }
        }
    }

    public void printFullSchema(PrintStream out) {
        print(out);
        out.format("\n");
        printTupleDefs(out);
        printInheritanceRules(out);
    }

    private void printTupleDefs(PrintStream out) {
        for (TableSchema ts : schemas) {
            ts.writeTupleDef(out);
        }
    }

    private void printInheritanceRules(PrintStream out) {
        out.format("\n");
        for (TableSchema ts : schemas) {
            ts.printInheritanceRule(out);
            SubTableSchema sts = findSubTableSchema(ts.getName());
            if (sts!=null)
                sts.printInheritanceRules(out);
        }
    }

    public void equals(DBSchema d) {
        if (!d.getName().equals(name)) {
            throw new Error("database schemas do not have the same name (" + d.getName() + "," + name + ")");
        }
        if (d.size() != size()) {
            throw new Error("database schemas do not have the same number of tables (" + d.getName() + "," + name + ")");
        }
        for (TableSchema ts : schemas) {
            ts.equals(d.findTableSchema(ts.getName())); // throws error otherwise
        }
    }
}
